==> emencia-crma-2.0.tar.gz: emencia-crma-2.0/crma/models.py <==
# -*- coding: utf-8 -*-

import hmac
import json
import datetime

from django.db import models, transaction
from django.db.models import Q
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils import six, timezone as tz
from django.utils.encoding import python_2_unicode_compatible
from django.utils.crypto import get_random_string
from django.core.exceptions import ObjectDoesNotExist


ST_SENT = 'sent'
ST_PENDING = 'pending'
ST_CANCELED = 'canceled'
ST_ERROR = 'error'


@python_2_unicode_compatible
class Channel(models.Model):
    """
    Contacts can subscribe to channels. Emails are attached to channels.  A
    contact will receive an email if he's subscribed to the email's channel.

    If the channel is marked as required (boolean field) then all contacts are
    implicitely subscribed to the channel. Typically there is one such channel,
    which could be called "Default", and can be used for sending administrative
    emails.
    """

    title = models.CharField(
        max_length=80,
    )
    # We use the channel_id to reference the channel in the code
    channel_id = models.CharField(
        max_length=80,
        unique=True,
    )
    from_address = models.CharField(
        max_length=200,
        help_text=_('"Name <email>" or "email"'),
    )
    required = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return self.title

    @transaction.atomic
    def schedule_or_update(self, contact, serialized_context=''):
        """
        Check a channel schedule. Create or update (date) the planification of
        emails. Is useful for channels of type reminder login
        """
        assert isinstance(serialized_context, six.text_types)
        assert isinstance(contact, Contact)

        subscription, created = Subscription.objects.get_or_create(
            contact_id=contact.pk,
            channel_id=self.pk,
        )
        if not subscription.is_active:
            return

        # Now we schedule or update the mails to send
        emails = Email.objects.filter(channel=self, enabled=True)

        now = datetime.datetime.now()
        for email in emails:
            planned_emails = EmailScheduler.objects.filter(
                email=email,
                status=ST_PENDING,
                contact=contact
            )
            try:
                email_scheduler = planned_emails[:1].get()
            except ObjectDoesNotExist:
                sched_time = now + email.interval
                EmailScheduler.objects.create(
                    email=email,
                    lang=contact.lang,
                    from_address=self.from_address,
                    contact=contact,
                    status=ST_PENDING,
                    raw_extra_context=serialized_context,
                    sched_time=sched_time,
                )
            else:
                email_scheduler.sched_time = now
                email_scheduler.save()

    @transaction.atomic
    def subscribe(self, contact, serialized_context=''):
        assert isinstance(serialized_context, six.string_types)
        assert isinstance(contact, Contact)

        subscription, created = Subscription.objects.get_or_create(
            contact_id=contact.pk,
            channel_id=self.pk,
        )
        if not subscription.is_active:
            subscription.state = Subscription.SUBSCRIBED
            subscription.save()

        # Now we Schedule the mails to send
        emails = Email.objects.filter(channel=self, enabled=True)
        for email in emails:
            sched_time = datetime.datetime.now() + email.interval
            EmailScheduler.objects.create(
                email=email,
                lang=contact.lang,
                from_address=self.from_address,
                contact=contact,
                status=ST_PENDING,
                raw_extra_context=serialized_context,
                sched_time=sched_time,
            )


@python_2_unicode_compatible
class Contact(models.Model):
    email = models.CharField(
        max_length=200,
    )
    lang = models.CharField(
        max_length=10,
        blank=True
    )
    subscriptions = models.ManyToManyField(
        Channel,
        through='Subscription',
    )

    def __str__(self):
        return self.email

    def generate_token(self):
        token = hmac.new(settings.SECRET_KEY.encode('utf-8'),
                         u'{}{}'.format(self.pk, self.email).encode('utf-8'))
        return token.hexdigest()


@python_2_unicode_compatible
class MailingList(models.Model):
    """
    A Mailing List is a list of members
    """
    title = models.CharField(
        max_length=90,
    )
    members = models.ManyToManyField(
        Contact,
        blank=True
    )
    all = models.BooleanField(
        default=False,
        help_text=_(u'If all is checked, the "members" field is ignored. '
                    u'Every contact in the database will be '
                    u'considered a member of this mailing list.')
    )

    def __str__(self):
        return self.title

    def get_contacts(self):
        if self.all:
            return Contact.objects.all()

        return self.members.all()

    @transaction.atomic
    def schedule(self, email, serialized_context='', plan_date=None):
        """
        Return the number of emails scheduled.
        """
        assert isinstance(email, Email)
        assert isinstance(serialized_context, six.string_types)

        count = 0
        for contact in self.get_contacts():
            obj = email.schedule(contact, serialized_context, plan_date)
            if obj is not None:
                count += 1

        return count


def generate_unsubscribe_key():
    return get_random_string()


class Subscription(models.Model):
    SUBSCRIBED = 'subscribed'
    UNSUBSCRIBED = 'unsubscribed'

    channel = models.ForeignKey(
        Channel,
    )
    contact = models.ForeignKey(
        Contact
    )
    state = models.CharField(
        max_length=20,
        choices=[
            (SUBSCRIBED, _('Subscribed')),
            (UNSUBSCRIBED, _('Unsubscribed')),
        ],
        default=SUBSCRIBED,
    )
    unsubscribe_key = models.CharField(
        _('Unsubscribe key'),
        max_length=40,
        default=generate_unsubscribe_key
    )

    @property
    def is_active(self):
        return self.state == self.SUBSCRIBED

    @transaction.atomic
    def cancel(self):
        EmailScheduler.objects.filter(
            contact=self.contact,
            email__channel=self.channel
        ).cancel_pending()
        self.state = Subscription.UNSUBSCRIBED
        self.save()


class Email(models.Model):
    """
    Email is an email template used to build emails for each contact
    """

    subject = models.CharField(
        _('Subject'),
        max_length=200,
    )
    body = models.TextField(
        blank=True,
    )
    interval = models.DurationField(
    )
    enabled = models.BooleanField(
    )
    channel = models.ForeignKey(
        Channel,
    )
    tag = models.CharField(
        max_length=30,
        blank=True
    )
    # We use the email_id to reference the email in the code
    email_id = models.CharField(
        max_length=50,
        unique=True,
    )

    def __str__(self):
        return self.subject

    @transaction.atomic
    def schedule(self, contact, serialized_context='', plan_date=None):
        assert isinstance(contact, Contact)

        if not self.enabled:
            return None

        subscription, created = Subscription.objects.get_or_create(
            contact_id=contact.pk,
            channel_id=self.channel_id
        )
        if subscription.state != Subscription.SUBSCRIBED:
            return None

        if plan_date is None:
            plan_date = datetime.datetime.now() + self.interval

        return EmailScheduler.objects.create(
            email=self,
            lang=contact.lang,
            from_address=self.channel.from_address,
            contact=contact,
            status=ST_PENDING,
            raw_extra_context=serialized_context,
            sched_time=plan_date,
        )


class EmailSchedulerQuerySet(models.QuerySet):
    def cancel_pending(self):
        return self.filter(status=ST_PENDING).update(status=ST_CANCELED)

    def sendable(self, now=None):
        now = now or tz.now()
        return self.filter(
            Q(key=get_crma_key()) | Q(key=''),
            status=ST_PENDING,
            sched_time__lt=now,
        )


def get_crma_key():
    return getattr(settings, 'CRMA_KEY', '')


class EmailScheduler(models.Model):
    """
    The schedule for an email. It links an Email instance for the template with
    a Contact instance for the target and a date.
    """
    objects = EmailSchedulerQuerySet.as_manager()

    ctime = models.DateTimeField(
        auto_now_add=True
    )
    sched_time = models.DateTimeField(
    )
    sent_time = models.DateTimeField(
        null=True,
        blank=True,
    )
    email = models.ForeignKey(
        Email,
    )
    lang = models.CharField(
        max_length=10,
        blank=True
    )
    from_address = models.CharField(
        max_length=200,
        null=True,
        blank=True,
    )
    contact = models.ForeignKey(
        Contact,
    )
    status = models.CharField(
        max_length=20,
        choices=[
            (ST_SENT, _('Sent')),
            (ST_PENDING, _('Pending')),
            (ST_CANCELED, _('Canceled')),
            (ST_ERROR, _('Error')),
        ],
        default=ST_SENT,
    )
    raw_extra_context = models.TextField(
        blank=True,
    )
    trace_error = models.TextField(
        blank=True,
    )

    def __repr__(self):
        return 'S[{}]({}=>{})'.format(self.pk, self.email, self.contact)

    # In the settings there is a CRMA_KEY, we only send the emails with
    # this key. It is a security check to prevent sending emails to users
    # when we import databases from / to production for example.
    key = models.CharField(
        max_length=80,
        blank=True,
        default=get_crma_key,
    )

    @property
    def extra_context(self):
        if not self.raw_extra_context:
            return {}
        return json.loads(self.raw_extra_context)

    def get_subscription(self):
        subs, created = Subscription.objects.get_or_create(
            contact_id=self.contact_id,
            channel_id=self.email.channel_id,
        )
        return subs

==> emencia-crma-2.2.0.tar.gz: emencia-crma-2.2.0/crma/models.py <==
# -*- coding: utf-8 -*-

import hmac
import json
import datetime

from django.db import models, transaction
from django.db.models import Q
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils import six, timezone as tz
from django.utils.encoding import python_2_unicode_compatible
from django.utils.crypto import get_random_string
from django.core.exceptions import ObjectDoesNotExist


ST_SENT = 'sent'
ST_PENDING = 'pending'
ST_CANCELED = 'canceled'
ST_ERROR = 'error'


@python_2_unicode_compatible
class Channel(models.Model):
    """
    Contacts can subscribe to channels. Emails are attached to channels.  A
    contact will receive an email if he's subscribed to the email's channel.

    If the channel is marked as required (boolean field) then all contacts are
    implicitely subscribed to the channel. Typically there is one such channel,
    which could be called "Default", and can be used for sending administrative
    emails.
    """

    title = models.CharField(
        max_length=80,
    )
    # We use the channel_id to reference the channel in the code
    channel_id = models.CharField(
        max_length=80,
        unique=True,
    )
    from_address = models.CharField(
        max_length=200,
        help_text=_('"Name <email>" or "email"'),
    )
    required = models.BooleanField(
        default=False,
    )

    def __str__(self):
        return self.title

    @transaction.atomic
    def schedule_or_update(self, contact, serialized_context=''):
        """
        Check a channel schedule. Create or update (date) the planification of
        emails. Is useful for channels of type reminder login
        """
        assert isinstance(serialized_context, six.string_types), '%r is not a string' % serialized_context
        assert isinstance(contact, Contact), '%r is not a Contact' % contact

        subscription, created = Subscription.objects.get_or_create(
            contact_id=contact.pk,
            channel_id=self.pk,
        )
        if not subscription.is_active:
            return

        # Now we schedule or update the mails to send
        emails = Email.objects.filter(channel=self, enabled=True)

        now = datetime.datetime.now()
        for email in emails:
            planned_emails = EmailScheduler.objects.filter(
                email=email,
                status=ST_PENDING,
                contact=contact
            )
            try:
                email_scheduler = planned_emails[:1].get()
            except ObjectDoesNotExist:
                sched_time = now + email.interval
                EmailScheduler.objects.create(
                    email=email,
                    lang=contact.lang,
                    from_address=self.from_address,
                    contact=contact,
                    status=ST_PENDING,
                    raw_extra_context=serialized_context,
                    sched_time=sched_time,
                )
            else:
                email_scheduler.sched_time = now
                email_scheduler.save()

    @transaction.atomic
    def subscribe(self, contact, serialized_context=''):
        assert isinstance(serialized_context, six.string_types), '%r is not a string' % serialized_context
        assert isinstance(contact, Contact), '%r is not a Contact' % contact

        subscription, created = Subscription.objects.get_or_create(
            contact_id=contact.pk,
            channel_id=self.pk,
        )
        if not subscription.is_active:
            subscription.state = Subscription.SUBSCRIBED
            subscription.save()

        # Now we Schedule the mails to send
        emails = Email.objects.filter(channel=self, enabled=True)
        for email in emails:
            sched_time = datetime.datetime.now() + email.interval
            EmailScheduler.objects.create(
                email=email,
                lang=contact.lang,
                from_address=self.from_address,
                contact=contact,
                status=ST_PENDING,
                raw_extra_context=serialized_context,
                sched_time=sched_time,
            )


@python_2_unicode_compatible
class Contact(models.Model):
    email = models.CharField(
        max_length=200,
    )
    lang = models.CharField(
        max_length=10,
        blank=True
    )
    subscriptions = models.ManyToManyField(
        Channel,
        through='Subscription',
    )

    def __str__(self):
        return self.email

    def generate_token(self):
        token = hmac.new(settings.SECRET_KEY.encode('utf-8'),
                         u'{}{}'.format(self.pk, self.email).encode('utf-8'))
        return token.hexdigest()


@python_2_unicode_compatible
class MailingList(models.Model):
    """
    A Mailing List is a list of members
    """
    title = models.CharField(
        max_length=90,
    )
    members = models.ManyToManyField(
        Contact,
        blank=True
    )
    all = models.BooleanField(
        default=False,
        help_text=_(u'If all is checked, the "members" field is ignored. '
                    u'Every contact in the database will be '
                    u'considered a member of this mailing list.')
    )

    def __str__(self):
        return self.title

    def get_contacts(self):
        if self.all:
            return Contact.objects.all()

        return self.members.all()

    @transaction.atomic
    def schedule(self, email, serialized_context='', plan_date=None):
        """
        Return the number of emails scheduled.
        """
        assert isinstance(email, Email), '%r is not an Email' % email
        assert isinstance(serialized_context, six.string_types), '%r is not a string' % serialized_context

        count = 0
        for contact in self.get_contacts():
            obj = email.schedule(contact, serialized_context, plan_date)
            if obj is not None:
                count += 1

        return count


def generate_unsubscribe_key():
    return get_random_string()


class Subscription(models.Model):
    SUBSCRIBED = 'subscribed'
    UNSUBSCRIBED = 'unsubscribed'

    channel = models.ForeignKey(
        Channel,
    )
    contact = models.ForeignKey(
        Contact
    )
    state = models.CharField(
        max_length=20,
        choices=[
            (SUBSCRIBED, _('Subscribed')),
            (UNSUBSCRIBED, _('Unsubscribed')),
        ],
        default=SUBSCRIBED,
    )
    unsubscribe_key = models.CharField(
        _('Unsubscribe key'),
        max_length=40,
        default=generate_unsubscribe_key
    )

    @property
    def is_active(self):
        return self.state == self.SUBSCRIBED

    def activate(self):
        self.state = self.SUBSCRIBED
        self.save(update_fields=['state'])

    @transaction.atomic
    def cancel(self):
        EmailScheduler.objects.filter(
            contact=self.contact,
            email__channel=self.channel
        ).cancel_pending()
        self.state = Subscription.UNSUBSCRIBED
        self.save()


def generate_email_id():
    return u'mail-{}'.format(get_random_string(8))


@python_2_unicode_compatible
class Email(models.Model):
    """
    Email is an email template used to build emails for each contact
    """

    subject = models.CharField(
        _('Subject'),
        max_length=200,
    )
    body = models.TextField(
        blank=True,
    )
    interval = models.DurationField(
    )
    enabled = models.BooleanField(
    )
    channel = models.ForeignKey(
        Channel,
    )
    tag = models.CharField(
        max_length=30,
        blank=True
    )
    # We use the email_id to reference the email in the code
    email_id = models.CharField(
        max_length=50,
        default=generate_email_id,
        unique=True,
    )

    def __str__(self):
        return self.subject

    @transaction.atomic
    def schedule(self, contact, serialized_context='', plan_date=None):
        assert isinstance(contact, Contact), '%r is not a Contact' % contact

        if not self.enabled:
            return None

        subscription, created = Subscription.objects.get_or_create(
            contact_id=contact.pk,
            channel_id=self.channel_id
        )
        if not subscription.is_active:
            return None

        if plan_date is None:
            plan_date = datetime.datetime.now() + self.interval

        return EmailScheduler.objects.create(
            email=self,
            lang=contact.lang,
            from_address=self.channel.from_address,
            contact=contact,
            status=ST_PENDING,
            raw_extra_context=serialized_context,
            sched_time=plan_date,
        )


class EmailSchedulerQuerySet(models.QuerySet):
    def cancel_pending(self):
        return self.filter(status=ST_PENDING).update(status=ST_CANCELED)

    def sendable(self, now=None):
        now = now or tz.now()
        return self.filter(
            Q(key=get_crma_key()) | Q(key=''),
            status=ST_PENDING,
            sched_time__lt=now,
        )


def get_crma_key():
    return getattr(settings, 'CRMA_KEY', '')


@python_2_unicode_compatible
class EmailScheduler(models.Model):
    """
    The schedule for an email. It links an Email instance for the template with
    a Contact instance for the target and a date.
    """
    objects = EmailSchedulerQuerySet.as_manager()

    ctime = models.DateTimeField(
        auto_now_add=True
    )
    sched_time = models.DateTimeField(
    )
    sent_time = models.DateTimeField(
        null=True,
        blank=True,
    )
    email = models.ForeignKey(
        Email,
    )
    lang = models.CharField(
        max_length=10,
        blank=True
    )
    from_address = models.CharField(
        max_length=200,
        null=True,
        blank=True,
    )
    contact = models.ForeignKey(
        Contact,
    )
    status = models.CharField(
        max_length=20,
        choices=[
            (ST_SENT, _('Sent')),
            (ST_PENDING, _('Pending')),
            (ST_CANCELED, _('Canceled')),
            (ST_ERROR, _('Error')),
        ],
        default=ST_SENT,
    )
    raw_extra_context = models.TextField(
        blank=True,
    )
    trace_error = models.TextField(
        blank=True,
    )

    def __str__(self):
        return u'{} to {}'.format(self.email, self.contact)

    def __repr__(self):
        return 'S[{}]({}=>{})'.format(self.pk, self.email, self.contact)

    # In the settings there is a CRMA_KEY, we only send the emails with
    # this key. It is a security check to prevent sending emails to users
    # when we import databases from / to production for example.
    key = models.CharField(
        max_length=80,
        blank=True,
        default=get_crma_key,
    )

    @property
    def extra_context(self):
        if not self.raw_extra_context:
            return {}
        return json.loads(self.raw_extra_context)

    def get_subscription(self):
        subs, created = Subscription.objects.get_or_create(
            contact_id=self.contact_id,
            channel_id=self.email.channel_id,
        )
        return subs
